var tetrisGame

(function(w, d){
"use strict"

var array2d = {
	width:0,
	length:0,
	array:null
}

array2d.create = function(array, length, width){
	if(array.length < length * width){
		throw new Error("array too small")
	}
	if(length < 0 || width < 0){
		throw new Error("no negative values for length or width")
	}
	
	var o = Object.create(this)
	o.width = width
	o.length = length
	o.array = array
	return o
}

array2d.index = function(row, column){
	if(row >= this.length || row < 0){
		throw new RangeError("array row index " + row + " out of bounds")
	}
	if(column >= this.width || column < 0){
		throw new RangeError("array column index " + column + " out of bounds")
	}
	
	return row * this.width + column
}

array2d.lookup = function(row, column){
	return this.array[this.index(row,column)]
}

array2d.isOutAt = function(row, column){
	return row >= this.length || row < 0 || column >= this.width || column < 0
}

array2d.isOutOrZeroAt = function(row, column){
	return this.isOutAt(row,column) || this.array[row * this.width + column] === 0
}

array2d.isOutOrNotZeroAt = function(row, column){
	return this.isOutAt(row,column) || this.array[row * this.width + column] !== 0
}



tetrisGame = {
	start:function(){},
	pause:function(){}
}

tetrisGame.create = function(){

//config
const size = 20
const borderSize = 1

const startInterval = 400
const shortInterval = 15
const speedIncrease = 0.2 //20%

const scorePerLineCleared = 100
const scorePerBlockSet = 10
const scorePerFastTetrisMove = 1

const blocks = [
	{color:"blue",bordercolor:"darkblue",size:3,startpos:0,shape:new Int32Array([
		1,1,1,0,
		0,0,1,0,
		0,0,0,0,
		0,0,0,0
	])},
	{color:"orange",bordercolor:"darkorange",size:3,startpos:0,shape:new Int32Array([
		1,1,1,0,
		1,0,0,0,
		0,0,0,0,
		0,0,0,0
	])},
	{color:"violet",bordercolor:"darkviolet",size:2,startpos:0,shape:new Int32Array([
		1,1,0,0,
		1,1,0,0,
		0,0,0,0,
		0,0,0,0
	])},
	{color:"green",bordercolor:"darkgreen",size:3,startpos:0,shape:new Int32Array([
		1,1,0,0,
		0,1,1,0,
		0,0,0,0,
		0,0,0,0
	])},
	{color:"red",bordercolor:"darkred",size:3,startpos:0,shape:new Int32Array([
		0,1,1,0,
		1,1,0,0,
		0,0,0,0,
		0,0,0,0
	])},
	{color:"turquoise",bordercolor:"darkturquoise",size:3,startpos:-1,shape:new Int32Array([
		0,0,0,0,
		1,1,1,0,
		0,1,0,0,
		0,0,0,0
	])},
	{color:"magenta",bordercolor:"darkmagenta",size:4,startpos:-1,shape:new Int32Array([
		0,0,0,0,
		1,1,1,1,
		0,0,0,0,
		0,0,0,0
	])}
] 


//initialization
var canvas = d.getElementById("canvas")
var c = canvas.getContext("2d")
c.lineWidth = borderSize

var tetrisSize = blocks[0].size
var tetrisIndex = 1
const tetrisRows = 4
const tetris = array2d.create(new Int32Array(tetrisRows * tetrisRows),tetrisRows,tetrisRows)

const hitboxWidth = canvas.width / size | 0
const hitboxLength = canvas.height / size | 0
const hitbox = array2d.create(new Int32Array(hitboxWidth * hitboxLength),hitboxLength,hitboxWidth)

// Detects if 'localStorage' is available or not, not all browsers support localstorage, fallback to cookies
var localstorageAvailable
try{
	window.localStorage
	localstorageAvailable = true
}catch(e){
	localstorageAvailable = false
}

var loading = true

var interval
var intervalId = 0
var isFast

var lines

var level
var levelScreen = d.getElementById("level")

var score = 0
var scoreScreen = d.getElementById("score")
var best
var bestScreen = d.getElementById("best")

var stopCounter = 0
var buttonPause = d.getElementById("pause")
var isPauseActive = true
var isPause = false

var buttonStart = d.getElementById("start")

var helpScreen = d.getElementById("help")
var isHelp = false

var gameover
var gameoverScreen = d.getElementById("gameover")

//x and y coordinate of the tetris block on the canvas
var x
var y


function clear(){
	var i,j
	for(i = 0;i < tetrisSize;i++){
		for(j = 0;j < tetrisSize;j++){
			if(tetris.lookup(i,j) == 1){
				c.clearRect(x + size * j,y + size * i,size,size)
			}
		}
	}
}

function clearArea(){
	var i
	c.clearRect(0,0,canvas.width,canvas.height)
	for(i = 0;i < hitbox.array.length;i++){
		hitbox.array[i] = 0
	}
}

function clearLine(){
	var i,j
	var height = 0
	var posY = y / size
	
	var sizeY = 0
	loop:for(i = tetris.length - 1;i >= 0;i--){
		for(j = 0;j < tetrisSize;j++){
			if(tetris.lookup(i,j) == 1){
				sizeY = i + 1
				break loop
			}
		}
	}
	
	for(i = sizeY - 1;i >= 0 && posY + i + height >= 0;i--){
		var isRow = true
		for(j = 0;j < hitbox.width;j++){
			if(hitbox.lookup(posY + i + height,j) == 0){
				isRow = false
				break
			}
		}
		if(isRow){
			height++
			updateScore(scorePerLineCleared)
		}
		if(height > 0 && posY + i > 0){
			for(j = 0;j < hitbox.width;j++){
				hitbox.array[hitbox.index(posY + i + height - 1,j)] = hitbox.lookup(posY + i - 1,j)
			}
		}
	}

	if(height == 0){
		return
	}

	for(i = posY - 1;i >= 0;i--){
		for(j = 0;j < hitbox.width;j++){
			hitbox.array[hitbox.index(i + height,j)] = hitbox.lookup(i,j)
		}
	}
	for(i = posY + sizeY - 1;i >= 0;i--){
		for(j = 0;j < hitbox.width;j++){
			var index = hitbox.lookup(i,j) - 1
			if(index != -1){
				c.fillStyle = blocks[index].color
				c.strokeStyle = blocks[index].bordercolor
				c.fillRect(size * j,size * i,size,size)
				c.strokeRect(size * j + borderSize,size * i + borderSize,size - 2 * borderSize,size - 2 * borderSize)
			}else{
				c.clearRect(size * j,size * i,size,size)
			}
		}
	}
	
	lines += height
	//update level by four new lines
	if(Math.floor(lines / 4) > level - 1){
		updateLevel()
	}
}

function draw(){
	var i,j
	for(i = 0;i < tetrisSize;i++){
		for(j = 0;j < tetrisSize;j++){
			if(tetris.lookup(i,j) == 1){
				c.fillRect(x + size * j,y + size * i,size,size)
				c.strokeRect(x + size * j + borderSize,y + size * i + borderSize,size - 2 * borderSize,size - 2 * borderSize)
			}
		}
	}
}

function moveSide(dir){
	var i,j
	var posX = x / size,posY = y / size

	//check if tetris block can move sidewards
	for(i = 0;i < tetrisSize;i++){
		for(j = 0;j < tetrisSize;j++){
			if(tetris.lookup(i,j) == 1 && hitbox.isOutOrNotZeroAt(posY + i,posX + j + dir)){
				return
			}
		}
	}
	
	clear()
	x += dir * size
	draw()
}

function move(){
	var i,j
	var posX = x / size,posY = y / size
	
	for(i = 0;i < tetrisSize;i++){
		for(j = tetrisSize - 1;j >= 0;j--){
			if(tetris.lookup(j,i) == 1 && hitbox.isOutOrNotZeroAt(posY + j + 1,posX + i)){
				//if hit
				for(i = 0;i < tetrisSize;i++){
					for(j = 0;j < tetrisSize;j++){
						if(tetris.lookup(i,j) == 1){
							//set location where tetris block is hit
							hitbox.array[hitbox.index(posY + i,posX + j)] = tetrisIndex
						}
					}
				}
				updateScore(scorePerBlockSet)
				clearLine()
				if(isFast){
					setMoveInterval(interval)
					isFast = false
				}
				newTetris()
				return
			}
		}
	}
	clear()
	y += size
	if(isFast){
		updateScore(scorePerFastTetrisMove)
	}
	draw()
}

function rotate(){
	var i,j
	var posX = x / size,posY = y / size
	var tetrisSize_2 = Math.floor(tetrisSize / 2)
	var previous
	
	//check if tetris block can rotate
	for(i = 0;i < tetrisSize_2;i++){
		var iInverse = tetrisSize - 1 - i
		for(j = 0;j < iInverse - i;j++){
			if((tetris.lookup(i + j,iInverse) == 1 && hitbox.isOutOrNotZeroAt(posY + i,posX + i + j)) ||
			   (tetris.lookup(iInverse,iInverse - j) == 1 && hitbox.isOutOrNotZeroAt(posY + i + j,posX + iInverse)) ||
			   (tetris.lookup(iInverse - j,i) == 1 && hitbox.isOutOrNotZeroAt(posY + iInverse,posX + iInverse - j)) ||
			   (tetris.lookup(i,i + j) == 1 && hitbox.isOutOrNotZeroAt(posY + iInverse - j,posX + i))){
				return
			}
		}
	}
	
	//rotate
	clear()
	for(i = 0;i < tetrisSize_2;i++){
		var iInverse = tetrisSize - 1 - i
		for(j = 0;j < iInverse - i;j++){
			previous = tetris.lookup(i,i + j)
			tetris.array[tetris.index(i,i + j)] = tetris.lookup(i + j,iInverse)
			tetris.array[tetris.index(i + j,iInverse)] = tetris.lookup(iInverse,iInverse - j)
			tetris.array[tetris.index(iInverse,iInverse - j)] = tetris.lookup(iInverse - j,i)
			tetris.array[tetris.index(iInverse - j,i)] = previous
		}
	}
	draw()
}

function skip(){
	var i,j,k
	var posX = x / size,posY = y / size
	var gap = hitbox.length
	
	for(i = 0;i < tetrisSize;i++){
		for(j = tetrisSize - 1;j >= 0;j--){
			if(tetris.lookup(j,i) == 1){
				for(k = j;;k++){
					if(hitbox.isOutOrNotZeroAt(posY + k + 1,posX + i)){
						if(gap > k - j){
							gap = k - j
						}
						break
					}
				}
			}
		}
	}
	
	clear()
	y += size * gap
	updateScore(gap * scorePerFastTetrisMove)
	draw()
	move()
}

function setMoveInterval(duration){
	if(intervalId){
		clearMoveInterval()
	}
	intervalId = w.setInterval(move,duration)
}

function clearMoveInterval(){
	w.clearInterval(intervalId)
	intervalId = 0
}

function keyDownHandler(e){
	if(stopCounter != 0){
		return
	}
	var k = e.keyCode
	switch(k){
	case 37: //arrow left
	case 39: //arrow right
		moveSide(k - 38)
		break
	case 38: //arrow up
		rotate()
		break
	case 40: //arrow down
		if(!isFast){
			setMoveInterval(shortInterval)
			isFast = true
		}
		break
	case 32: //space
		skip()
	}
}

function keyUpHandler(e){
	if(stopCounter != 0){
		return
	}
	if(e.keyCode == 40){ //arrow down
		setMoveInterval(interval)
		isFast = false
	}
}

function newTetris(){
	var i,j
	var rng = Math.random() * blocks.length | 0
	tetris.array.set(blocks[rng].shape)
	tetrisSize = blocks[rng].size
	tetrisIndex = rng + 1
	
	var posY = blocks[rng].startpos
	var posX = Math.floor(canvas.width / size / 2) - Math.floor(tetrisSize / 2)
	y = posY * size
	x = posX * size
	for(i = -posY;i < tetrisSize;i++){
		for(j = 0;j < tetrisSize;j++){
			if(tetris.lookup(i,j) == 1 && hitbox.isOutOrNotZeroAt(i + posY,posX + j)){
				if(score > best){
					best = score
					saveBestScore()
				}
				//alert("GAME OVER")
				displayGameOver()
				return
			}
		}
	}
	c.fillStyle = blocks[rng].color
	c.strokeStyle = blocks[rng].bordercolor
	draw()
}

function updateScore(value){
	score += value
	scoreScreen.textContent = score
}

function showLevel(){
	levelScreen.textContent = level
}

function updateLevel(){
	level++
	interval = startInterval / (1 + level * speedIncrease)
	w.clearInterval(intervalId)
	intervalId = w.setInterval(move,interval)
	showLevel()
}

function saveBestScore(){
	if(localstorageAvailable){
		w.localStorage.setItem("score", score)
	}else{
		// old cookie storage, localStorage suits better for browser data storage
		d.cookie = "score=" + score + ";max-age=" + 20 * 31536e3 //20 years
	}
	bestScreen.textContent = best
}

function retrieveBestScore(){
	if(localstorageAvailable){
		best = parseInt(w.localStorage.getItem("score"))
	}else{
		// old cookie storage, localStorage suits better for browser data storage
		best = parseInt(d.cookie.replace(/(?:(?:^|.*;\s*)score\s*\=\s*([^;]*).*$)|^.*$/, "$1"),10)
	}
	
	if(isNaN(best)){
		best = 0
	}
	bestScreen.textContent = best
}

function startOrResetBestScore(e){
	if(isHelp){ //start button deactivated when help menu is appearing
		return
	}
	if(e.shiftKey){ //"SHIFT"
		best = 0
		saveBestScore()
	}else{
		start()
	}
}

function stop(){
	if(stopCounter == 0){
		clearMoveInterval()
	}
	stopCounter++
}

function resume(){
	if(stopCounter > 0){
		stopCounter--
	}
	if(stopCounter == 0){
		setMoveInterval(interval)
	}
}

function start(){
	score = 0
	lines = 0
	level = 1
	showLevel()
	updateScore(0)
	retrieveBestScore()
	interval = startInterval
	isFast = false
	if(gameover){
		//leave gameover screen
		displayGameOver()
	}
	if(!loading){
		clearArea()
	}else{
		loading = false
	}
	newTetris()
	isPauseActive = true
	isPause = true
	pause()
}

function pause(){
	if(!isPauseActive){
		return
	}
	if(!isPause){
		stop()
	}else{
		resume()
	}
	buttonPause.textContent = !isPause ? "Continue" : "Pause"	
	isPause = !isPause
}

function startPauseHandler(e){
	switch(e.keyCode){
		case 114: //"R"
		case 82: //"SHIFT + R"
			startOrResetBestScore(e)
			return
		case 112: //"P"
			pause()
			return
	}
}


function displayHelp(e){
	var k = e.keyCode
	
	if(!isHelp){
		if(k != 112){ //"F1"
			return
		}
		stop()
		isPauseActive = false
	}else{
		if(k != 112 && k != 27){ //"F1","ESC"
			return
		}
		resume()
		isPauseActive = true
	}
	//no browser help
	e.preventDefault()
	helpScreen.classList.toggle("hidden")
	isHelp = !isHelp
}

function displayGameOver(e){
	if(!gameover){
		w.addEventListener("keyup",displayGameOver)
		stop()
		isPauseActive = false
	}else{
		if(e !== undefined && e.keyCode != 27){ //"ESC"
			return
		}
		w.removeEventListener("keyup",displayGameOver)
	}
	gameoverScreen.classList.toggle("hidden")
	gameover = !gameover
}

w.addEventListener("keydown",displayHelp)
w.addEventListener("keydown",keyDownHandler)
w.addEventListener("keyup",keyUpHandler)
w.addEventListener("keypress",startPauseHandler)
buttonPause.addEventListener("click",pause)
buttonStart.addEventListener("click",startOrResetBestScore)

//internet explorer
if("onhelp" in w){
	w.addEventListener("help",function(e){
		e.preventDefault()
	})
}


var o = Object.create(this)
o.start = start //not startOrResetBestScore, 'start' is all it needs for the beginning
o.pause = pause
return o

}

})(window,document)